!Program for inverting the Full Newton FWI Hessian matrix with MKL RCI CG library
program FN_inversion

	use sep
	use Hess_full_mod_1
	use dot_mod
	implicit none
	include 'mkl_rci.fi'
	
	integer 							:: nz, nx, ny, rc, sreed, wc, srite
	real								:: dz, dx, dy, oz, ox, oy, rnorm
	double precision					:: normalize
	real, dimension(:),allocatable		:: model_in, model_out, gradient
	!MKL RCI CG variables
	!We want to invert Hx = b
	integer								:: n 			! model size
	double precision, allocatable		:: x(:) 		! approximate solution
	double precision, allocatable		:: b(:)			! right-hand side of equation 
	integer								:: RCI_request 	! variable for handling RCs
	integer								:: ipar(128) 	! Integer parameters
	double precision					:: dpar(128) 	! Double precision parameters
	double precision, allocatable		:: tmp(:,:)		! temporary array for computation
	integer								:: itercount, niter
	
	call sep_init()
	
	call from_history("n1", nz)
  	call from_history("o1", oz)
 	call from_history("d1", dz)
  	call from_history("n2", nx)
  	call from_history("o2", ox)
  	call from_history("d2", dx)
  	call from_history("n3", ny)
  	call from_history("o3", oy)
    call from_history("d3", dy)
    call from_par("niter",niter)
    
    
    write(0,*)"Full Newton Hessian inversion with MKL RCI CG routine"
  	
  	!Initialize arrays for inversion
  	n = nz*nx*ny ! model size
  	allocate(x(n),b(n)); x = 0.; b = 0.
  	allocate(tmp(n,3)); tmp = 0.
  	
  	
  	!allocate memory for Hessian application
  	allocate(model_in(nz*nx*ny),model_out(nz*nx*ny),gradient(nz*nx*ny))
	!reading the gradient
	rc = sreed("in",gradient,4*nz*nx*ny)
	b = -gradient !Inverting the system H m = -g
	
	!writing history for residual and norm(r)
	call to_history("n1", nz,"residual")
  	call to_history("o1", oz,"residual")
 	call to_history("d1", dz,"residual")
  	call to_history("n2", nx,"residual")
  	call to_history("o2", ox,"residual")
  	call to_history("d2", dx,"residual")
  	call to_history("n3", ny,"residual")
  	call to_history("o3", oy,"residual")
  	call to_history("d3", dy,"residual")
  	call to_history("o4", 1., "residual")
  	call to_history("d4", 1., "residual")
  	
  	call to_history("n2", 1, "rnorm")
  	call to_history("o1", 1., "rnorm")
  	call to_history("d1", 1., "rnorm")
  	
  	!History for output
  	call to_history("n1", nz)
  	call to_history("o1", oz)
 	call to_history("d1", dz)
  	call to_history("n2", nx)
  	call to_history("o2", ox)
  	call to_history("d2", dx)
  	call to_history("n3", ny)
  	call to_history("o3", oy)
  	call to_history("d3", dy)
  	
	
  	
  	
  	write(0,*)"Initialize routines"
  	!Gauss-Newton Hessian computation routines
  	call Hess_full_init(nz, nx, ny)
  	!Initialize MKL inversion routine
  	call dcg_init(n, x, b, RCI_request, ipar, dpar, tmp)
  	!Change parameters to your inversion flavor
  	
  	!Check variables for inversion
  	call dcg_check(n, x, b, RCI_request, ipar, dpar, tmp)
  	if (RCI_request .ne. 0 ) call erexit("Not passing dcg_check")
	!If the starting model is equal zero
	itercount = -1

  	
  	!Running the solver
  	do 
  		call dcg(n, x, b, RCI_request, ipar, dpar, tmp)
  		if (RCI_request.eq.0) then !Apply the operator
  			write(0,*)"Solution found"
  			exit	
  		elseif(RCI_request.eq.1)then
			if(itercount .ge. 0)then
  				write(0,*)"Applying the Hessian"
  				model_in = tmp(1:n,1)
  				call Hess_full_call(.false., .false., model_in, model_out)
  				tmp(1:n,2) = model_out
			endif
  		elseif(RCI_request.eq.2) then !Stop test
  			call dcg_get(n, x, b, RCI_request, ipar, dpar, tmp, itercount)
  			!writing residual and norm of the residual
  			if(itercount .ne. 0) then
  				call to_history("n4", itercount,"residual")
				call to_history("n4", itercount)
  				call to_history("n1", itercount,"rnorm")
  				gradient = tmp(:,3) !converting to single precision
  				if(itercount .eq. 1) normalize = sum(tmp(:,3)*tmp(:,3))
  				rnorm = sum(tmp(:,3)*tmp(:,3))/normalize*100.0
  				write(0,*)"Residual norm percentage:",rnorm
  				wc = srite("residual",gradient,4*nz*nx*ny)
  				wc = srite("rnorm",rnorm,4)
				!writing approximate solution
				gradient = x; !converting to single precision
				wc = srite("out",gradient,4*nz*nx*ny)
  			endif
  			write(0,*)"Checking for stop, at iter:", itercount
  			if(itercount .ge. niter) then
  				write(0,*)"Max number of iterations reached: Stopping solver"
  				exit
  			endif
  		elseif(RCI_request.eq.3) then ! preconditioner
  			write(0,*)"Applying preconditioner (if present)"
  			cycle
  		elseif(RCI_request.lt.0) then
  			write(0,*)"Error in running the inversion: negative flag from solver",RCI_request
  			exit
  		else
  			write(0,*)"Unknown flag: stopping"
  			exit
  		endif
  	enddo
  	
	
	
	!Close Hessian computation
	call Hess_full_close()
	
  	
  	
  	deallocate(gradient,model_out,model_in)
  	deallocate(tmp)
  	deallocate(b,x)
  	call sep_close()
  	call exit (0)

end program
