!Module for applying Gauss-Newton FWI Hessian (i.e., (Born*)(Born))
!Takes a non-extended model, runs Born followed by RTM, and return a non-extended model
module Hess_GN_mod

  use sep

  implicit none
  !Module variables
  character*55,private							:: wavelet=' ',sources=' ',recs=' ',wavefields=' ',wantwavefield=' ',readwavefield=' ',wavefield=' ',dt=' ',threads=' ',background_vel=' ',padded=' '
  character*1024,private						:: command
  integer,private								:: nz, nx, ny
  real, dimension(:,:,:), allocatable, private	:: model_tmp
  
  
  contains
	
	!Initialization routine
	subroutine Hess_GN_init(nz_in,nx_in,ny_in)!,background_vel_in,wavelet_in,sources_in,recs_in,wavefields_in,wantwavefield_in,readwavefield_in,wavefield_in,dt_in,threads_in,padded_in)
		
		implicit none
 		character*40 ::wavelet_in,sources_in,recs_in,wavefields_in,wantwavefield_in,readwavefield_in,background_vel_in,dt_in,threads_in,wavefield_in,padded_in
		integer					:: nz_in,nx_in,ny_in

		!Initialize all the variables
		
!		write(background_vel,*)'<',trim(adjustl(background_vel_in)),' '
!		write(wavelet,*)'wavelet=',trim(adjustl(wavelet_in)),' '
!		write(sources,*)'sources=',trim(adjustl(sources_in)),' '
!		write(recs,*)'recs=',trim(adjustl(recs_in)),' '
!		write(wavefields,*)'wavefields=',trim(adjustl(wavefields_in)),' '
!		write(wantwavefield,*)'wantwavefield=',trim(adjustl(wantwavefield_in)),' '
!		write(readwavefield,*)'readwavefield=',trim(adjustl(readwavefield_in)),' '
!		write(wavefield,*)'wavefield=',trim(adjustl(wavefield_in)),' '
!		write(dt,*)'dt=',trim(adjustl(dt_in)),' ' 
!		write(threads,*)'threads=',trim(adjustl(threads_in)),' '
!		write(padded,*)'padded=',trim(adjustl(padded_in)),' '
		nz = nz_in
		nx = nx_in
		ny = ny_in
		allocate(model_tmp(nz,nx,ny)); model_tmp = 0.
		
		call sep_init()

		return
	end subroutine Hess_GN_init

  	subroutine Hess_GN_call(adj,add,mm,dd)
    	logical,intent(in)   	:: adj,add
   		real,dimension(:)  		:: mm
   		real,dimension(:)    	:: dd
    		call Hess_GN_op(adj,add,mm,dd)
  	end subroutine Hess_GN_call

	
	!Operator routine
	subroutine Hess_GN_op(adj, add, mm1, mm2)
	
		!Input-Output variables
		logical, intent(in)								:: adj, add
		real, dimension(nz,nx,ny), intent(inout)		:: mm1 
		real, dimension(nz,nx,ny), intent(inout)		:: mm2 
		!Internal variables
		integer											:: wc, srite, rc, sreed, sk, sseek
		
    	if(adj) then !Adjoint=Forward
        	if (.not. add) mm1 = 0. !Zero-out the model
        	wc=srite("tmp/model_in.H",mm2,nz*nx*ny*4)
			command = 'bin/Bornfast_fw.x <background.H wavelet=wavelet.H deltavel=tmp/model_in.H sources=Par/sources.p recs=Par/receivers.p wavefields=wavefield_fw_background.H readwavefield=0 wantwavefield=0 dt=1e-3 padded=1 threads=16 >tmp/born_data_tmp.H'
			call system(command)
			command = 'bin/Bornfast_adj.x <background.H wavelet=wavelet.H deltadata=tmp/born_data_tmp.H sources=Par/sources.p recs=Par/receivers.p  wavefields=wavefield_fw_background.H readwavefield=0 dt=1e-3 padded=1 threads=16 >tmp/born_model_tmp.H'
			call system(command)
			call system('Transp < tmp/born_model_tmp.H reshape=2,4 | Stack | Window n1=1 >tmp/model_tmp.H out=tmp/model_tmp.H@')
			rc = sreed('tmp/model_tmp.H',model_tmp,nz*nx*ny*4)
			sk = sseek("tmp/model_in.H",0,0)
			sk = sseek("tmp/model_tmp.H",-4*nz*nx*ny,1)
			mm1 = mm1 + model_tmp
			

		else	     !Forward
			if(.not. add) mm2 = 0. !Zero-out the data	
			wc=srite("tmp/model_in.H",mm1,nz*nx*ny*4)
			!command = 'bin/Bornfast_fw.x ' // background_vel // wavelet // 'deltavel=tmp/model_in.H' // sources // recs // wantwavefield // wavefields // readwavefield // wavefield // dt // padded // threads // '>tmp/born_data_tmp.H'
			command = 'bin/Bornfast_fw.x <background.H wavelet=wavelet.H deltavel=tmp/model_in.H sources=Par/sources.p recs=Par/receivers.p wavefields=wavefield_fw_background.H readwavefield=0 wantwavefield=0 dt=1e-3 padded=1 threads=16 >tmp/born_data_tmp.H verbose=1'
			!write(command,"(289A,$)")'bin/Bornfast_fw.x ',trim(adjustl(background_vel)),' ',trim(adjustl(wavelet)),' ','deltavel=tmp/model_in.H',' ',trim(adjustl(sources)),' ',trim(adjustl(recs)),' ',trim(adjustl(wantwavefield)),' ',trim(adjustl(wavefields)),' ' ,trim(adjustl(readwavefield)),' ',trim(adjustl(wavefield)),' ',trim(adjustl(dt)),' ',trim(adjustl(padded)),' ',trim(adjustl(threads)),' ','>tmp/born_data_tmp.H ',' verbose=1';
			call system(command)
			command = 'bin/Bornfast_adj.x <background.H wavelet=wavelet.H deltadata=tmp/born_data_tmp.H sources=Par/sources.p recs=Par/receivers.p  wavefields=wavefield_fw_background.H readwavefield=0 dt=1e-3 padded=1 threads=16 >tmp/born_model_tmp.H verbose=1'
			call system(command)
			call system('Transp < tmp/born_model_tmp.H reshape=2,4 | Stack | Window n1=1 >tmp/model_tmp.H out=tmp/model_tmp.H@')
			rc = sreed('tmp/model_tmp.H',model_tmp,nz*nx*ny*4)
			mm2 = mm2 + model_tmp
			sk = sseek("tmp/model_in.H",0,0)
			sk = sseek("tmp/model_tmp.H",-4*nz*nx*ny,1)
				
		endif
		
		return
	end subroutine Hess_GN_op
	
	subroutine Hess_GN_close()
		if(allocated(model_tmp)) deallocate(model_tmp)
	end subroutine Hess_GN_close
	

end module Hess_GN_mod