!Module for applying Gauss-Newton FWI Hessian (i.e., (Born*)(Born))
!Takes a non-extended model, runs Born followed by RTM, and return a non-extended model
module Hess_GN_mod_2

  use sep

  implicit none
  !Module variables
  character*55,private							:: wavelet=' ',sources=' ',recs=' ',wavefields=' ',wantwavefield=' ',readwavefield=' ',wavefield=' ',dt=' ',threads=' ',background_vel=' ',padded=' '
  character*1024,private						:: command
  integer,private								:: nz, nx, ny
  real, dimension(:,:,:), allocatable, private	:: model_tmp
  
  
  contains
	
	!Initialization routine
	subroutine Hess_GN_init(nz_in,nx_in,ny_in)
		
		implicit none
 		character*40 ::wavelet_in,sources_in,recs_in,wavefields_in,wantwavefield_in,readwavefield_in,background_vel_in,dt_in,threads_in,wavefield_in,padded_in
		integer					:: nz_in,nx_in,ny_in

		!Initialize all the variables
		nz = nz_in
		nx = nx_in
		ny = ny_in
		allocate(model_tmp(nz,nx,ny)); model_tmp = 0.
		
		call sep_init()

		return
	end subroutine Hess_GN_init

  	subroutine Hess_GN_call(adj,add,mm,dd)
    	logical,intent(in)   	:: adj,add
   		real,dimension(:)  		:: mm
   		real,dimension(:)    	:: dd
    		call Hess_GN_op(adj,add,mm,dd)
  	end subroutine Hess_GN_call

	
	!Operator routine
	subroutine Hess_GN_op(adj, add, mm1, mm2)
	
		!Input-Output variables
		logical, intent(in)								:: adj, add
		real, dimension(nz,nx,ny), intent(inout)		:: mm1 
		real, dimension(nz,nx,ny), intent(inout)		:: mm2 
		!Internal variables
		integer											:: wc, srite, rc, sreed, sk, sseek
		
    	if(adj) then !Adjoint=Forward
        	if (.not. add) mm1 = 0. !Zero-out the model
        	wc=srite("tmp/model_in.H",mm2,nz*nx*ny*4)
			command = 'bin/Bornfast_fw.x <true_background.H wavelet=wavelet.H deltavel=tmp/model_in.H sources=Par/sources.p recs=Par/receivers.p wavefields=wavefield_fw_background.H readwavefield=0 wantwavefield=0 dt=1e-3 padded=1 threads=16 >tmp/born_data_tmp.H'
			call system(command)
			command = 'bin/Bornfast_adj.x <true_background.H wavelet=wavelet.H deltadata=tmp/born_data_tmp.H sources=Par/sources.p recs=Par/receivers.p  wavefields=wavefield_fw_background.H readwavefield=0 dt=1e-3 padded=1 threads=16 >tmp/born_model_tmp.H'
			call system(command)
			call system('Transp < tmp/born_model_tmp.H reshape=2,4 | Stack | Window n1=1 >tmp/model_tmp.H out=tmp/model_tmp.H@')
			rc = sreed('tmp/model_tmp.H',model_tmp,nz*nx*ny*4)
			sk = sseek("tmp/model_in.H",0,0)
			sk = sseek("tmp/model_tmp.H",-4*nz*nx*ny,1)
			mm1 = mm1 + model_tmp
			

		else	     !Forward
			if(.not. add) mm2 = 0. !Zero-out the data	
			wc=srite("tmp/model_in.H",mm1,nz*nx*ny*4)
			command = 'bin/Bornfast_fw.x <true_background.H wavelet=wavelet.H deltavel=tmp/model_in.H sources=Par/sources.p recs=Par/receivers.p wavefields=wavefield_fw_background.H readwavefield=0 wantwavefield=0 dt=1e-3 padded=1 threads=16 >tmp/born_data_tmp.H verbose=1'
			call system(command)
			command = 'bin/Bornfast_adj.x <true_background.H wavelet=wavelet.H deltadata=tmp/born_data_tmp.H sources=Par/sources.p recs=Par/receivers.p  wavefields=wavefield_fw_background.H readwavefield=0 dt=1e-3 padded=1 threads=16 >tmp/born_model_tmp.H verbose=1'
			call system(command)
			call system('Transp < tmp/born_model_tmp.H reshape=2,4 | Stack | Window n1=1 >tmp/model_tmp.H out=tmp/model_tmp.H@')
			rc = sreed('tmp/model_tmp.H',model_tmp,nz*nx*ny*4)
			mm2 = mm2 + model_tmp
			sk = sseek("tmp/model_in.H",0,0)
			sk = sseek("tmp/model_tmp.H",-4*nz*nx*ny,1)
				
		endif
		
		return
	end subroutine Hess_GN_op
	
	subroutine Hess_GN_close()
		if(allocated(model_tmp)) deallocate(model_tmp)
	end subroutine Hess_GN_close
	

end module Hess_GN_mod_2